"""
A looping construct for a workflow.
There are multiple types of loop,
notably looping over a set of values or until a condition holds.
"""

from __future__ import annotations

from collections import defaultdict
import copy
from pprint import pp
import pprint
from typing import Dict, List, Optional, Tuple, Union, Any
from warnings import warn
from collections import defaultdict
from itertools import chain
from typing import cast, TYPE_CHECKING
from typing_extensions import override

from hpcflow.sdk.core.app_aware import AppAware
from hpcflow.sdk.core.actions import EARStatus
from hpcflow.sdk.core.skip_reason import SkipReason
from hpcflow.sdk.core.errors import LoopTaskSubsetError
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.loop_cache import LoopCache, LoopIndex
from hpcflow.sdk.core.enums import InputSourceType, TaskSourceType
from hpcflow.sdk.core.utils import check_valid_py_identifier, nth_key, nth_value
from hpcflow.sdk.utils.strings import shorten_list_str
from hpcflow.sdk.log import TimeIt

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from typing import Any, ClassVar
    from typing_extensions import Self, TypeIs
    from rich.status import Status
    from ..typing import DataIndex, ParamSource
    from .parameters import SchemaInput, InputSource
    from .rule import Rule
    from .task import WorkflowTask
    from .types import IterableParam
    from .workflow import Workflow, WorkflowTemplate


# @dataclass
# class StoppingCriterion:
#     parameter: Parameter
#     condition: ConditionLike


# @dataclass
# class Loop:
#     parameter: Parameter
#     stopping_criteria: StoppingCriterion
#     # TODO: should be a logical combination of these (maybe provide a superclass in valida to re-use some logic there?)
#     maximum_iterations: int


class Loop(JSONLike):
    """
    A loop in a workflow template.

    Parameters
    ----------
    tasks: list[int | ~hpcflow.app.WorkflowTask]
        List of task insert IDs or workflow tasks.
    num_iterations:
        Number of iterations to perform.
    name: str
        Loop name.
    non_iterable_parameters: list[str]
        Specify input parameters that should not iterate.
    termination: v~hpcflow.app.Rule
        Stopping criterion, expressed as a rule.
    termination_task: int | ~hpcflow.app.WorkflowTask
        Task at which to evaluate the termination condition.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(name="termination", class_name="Rule"),
    )

    @classmethod
    def __is_WorkflowTask(cls, value) -> TypeIs[WorkflowTask]:
        return isinstance(value, cls._app.WorkflowTask)

    def __init__(
        self,
        tasks: Iterable[int | str | WorkflowTask],
        num_iterations: int,
        name: str | None = None,
        non_iterable_parameters: list[str] | None = None,
        termination: Rule | None = None,
        termination_task: int | str | WorkflowTask | None = None,
    ) -> None:
        _task_refs: list[int | str] = [
            ref_i.insert_ID if self.__is_WorkflowTask(ref_i) else ref_i for ref_i in tasks
        ]

        if termination_task is None:
            _term_task_ref = _task_refs[-1]  # terminate on final task by default
        elif self.__is_WorkflowTask(termination_task):
            _term_task_ref = termination_task.insert_ID
        else:
            _term_task_ref = termination_task

        if _term_task_ref not in _task_refs:
            raise ValueError(
                f"If specified, `termination_task` (provided: {termination_task!r}) must "
                f"refer to a task that is part of the loop. Available task references "
                f"are: {_task_refs!r}."
            )

        self._task_refs = _task_refs
        self._task_insert_IDs = (
            cast("list[int]", _task_refs)
            if all(isinstance(ref_i, int) for ref_i in _task_refs)
            else None
        )

        self._num_iterations = num_iterations
        self._name = check_valid_py_identifier(name) if name else name
        self._non_iterable_parameters = non_iterable_parameters or []
        self._termination = termination
        self._termination_task_ref = _term_task_ref
        self._termination_task_insert_ID = (
            _term_task_ref if isinstance(_term_task_ref, int) else None
        )

        self._workflow_template: WorkflowTemplate | None = (
            None  # assigned by parent WorkflowTemplate
        )

    @override
    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        out = super()._postprocess_to_dict(d)
        return {k.lstrip("_"): v for k, v in out.items()}

    @classmethod
    def _json_like_constructor(cls, json_like: dict) -> Self:
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        if "task_insert_IDs" in json_like:
            insert_IDs = json_like.pop("task_insert_IDs")
        else:
            insert_IDs = json_like.pop("tasks")  # e.g. from YAML

        if "termination_task_insert_ID" in json_like:
            tt_iID = json_like.pop("termination_task_insert_ID")
        elif "termination_task" in json_like:
            tt_iID = json_like.pop("termination_task")  # e.g. from YAML
        else:
            tt_iID = None

        task_refs = None
        term_task_ref = None
        if "task_refs" in json_like:
            task_refs = json_like.pop("task_refs")
        if "termination_task_ref" in json_like:
            term_task_ref = json_like.pop("termination_task_ref")

        obj = cls(tasks=insert_IDs, termination_task=tt_iID, **json_like)
        if task_refs:
            obj._task_refs = task_refs
        if term_task_ref is not None:
            obj._termination_task_ref = term_task_ref

        return obj

    @property
    def task_refs(self) -> tuple[int | str, ...]:
        """Get the list of task references (insert IDs or task unique names) that define
        the extent of the loop."""
        return tuple(self._task_refs)

    @property
    def task_insert_IDs(self) -> tuple[int, ...]:
        """Get the list of task insert_IDs that define the extent of the loop."""
        assert self._task_insert_IDs
        return tuple(self._task_insert_IDs)

    @property
    def name(self) -> str | None:
        """
        The name of the loop, if one was provided.
        """
        return self._name

    @property
    def num_iterations(self) -> int:
        """
        The number of loop iterations to do.
        """
        return self._num_iterations

    @property
    def non_iterable_parameters(self) -> Sequence[str]:
        """
        Which parameters are not iterable.
        """
        return self._non_iterable_parameters

    @property
    def termination(self) -> Rule | None:
        """
        A termination rule for the loop, if one is provided.
        """
        return self._termination

    @property
    def termination_task_ref(self) -> int | str:
        """
        The unique name of the task at which the loop will terminate.
        """
        return self._termination_task_ref

    @property
    def termination_task_insert_ID(self) -> int:
        """
        The insert ID of the task at which the loop will terminate.
        """
        assert self._termination_task_insert_ID is not None
        return self._termination_task_insert_ID

    @property
    def termination_task(self) -> WorkflowTask:
        """
        The task at which the loop will terminate.
        """
        if (wt := self.workflow_template) is None:
            raise RuntimeError(
                "Workflow template must be assigned to retrieve task objects of the loop."
            )
        assert wt.workflow
        return wt.workflow.tasks.get(insert_ID=self.termination_task_insert_ID)

    @property
    def workflow_template(self) -> WorkflowTemplate | None:
        """
        The workflow template that contains this loop.
        """
        return self._workflow_template

    @workflow_template.setter
    def workflow_template(self, template: WorkflowTemplate):
        self._workflow_template = template

    def __workflow(self) -> None | Workflow:
        if (wt := self.workflow_template) is None:
            return None
        return wt.workflow

    @property
    def task_objects(self) -> tuple[WorkflowTask, ...]:
        """
        The tasks in the loop.
        """
        if not (wf := self.__workflow()):
            raise RuntimeError(
                "Workflow template must be assigned to retrieve task objects of the loop."
            )
        return tuple(wf.tasks.get(insert_ID=t_id) for t_id in self.task_insert_IDs)

    def _validate_against_workflow(self, workflow: Workflow) -> None:
        """Validate the loop parameters against the associated workflow."""

        names = workflow.get_task_unique_names(map_to_insert_ID=True)
        if self._task_insert_IDs is None:
            err = lambda ref_i: (
                f"Loop {self.name + ' ' if self.name else ''}has an invalid "
                f"task reference {ref_i!r}. Such a task does not exist in "
                f"the associated workflow, which has task names/insert IDs: "
                f"{names!r}."
            )
            self._task_insert_IDs = []
            for ref_i in self._task_refs:
                if isinstance(ref_i, str):
                    # reference is task unique name, so retrieve the insert ID:
                    try:
                        iID = names[ref_i]
                    except KeyError:
                        raise ValueError(err(ref_i)) from None
                else:
                    try:
                        workflow.tasks.get(insert_ID=ref_i)
                    except ValueError:
                        raise ValueError(err(ref_i)) from None
                    iID = ref_i
                self._task_insert_IDs.append(iID)

        if self._termination_task_insert_ID is None:
            tt_ref = self._termination_task_ref
            if isinstance(tt_ref, str):
                tt_iID = names[tt_ref]
            else:
                tt_iID = tt_ref
            self._termination_task_insert_ID = tt_iID

    def __repr__(self) -> str:
        num_iterations_str = ""
        if self.num_iterations is not None:
            num_iterations_str = f", num_iterations={self.num_iterations!r}"

        name_str = ""
        if self.name:
            name_str = f", name={self.name!r}"

        return (
            f"{self.__class__.__name__}("
            f"task_insert_IDs={self.task_insert_IDs!r}{num_iterations_str}{name_str}"
            f")"
        )

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        kwargs = self.to_dict()
        kwargs["tasks"] = kwargs.pop("task_insert_IDs")
        kwargs["termination_task"] = kwargs.pop("termination_task_insert_ID")
        task_refs = kwargs.pop("task_refs", None)
        term_task_ref = kwargs.pop("termination_task_ref", None)
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._workflow_template = self._workflow_template
        obj._task_refs = task_refs
        obj._termination_task_ref = term_task_ref
        return obj


class WorkflowLoop(AppAware):
    """
    Class to represent a :py:class:`.Loop` that is bound to a
    :py:class:`~hpcflow.app.Workflow`.

    Parameters
    ----------
    index: int
        The index of this loop in the workflow.
    workflow: ~hpcflow.app.Workflow
        The workflow containing this loop.
    template: Loop
        The loop that this was generated from.
    num_added_iterations:
        Description of what iterations have been added.
    iterable_parameters:
        Description of what parameters are being iterated over.
    output_parameters:
        Description of what parameter are output from this loop, and the final task insert
        ID from which they are output.
    parents: list[str]
        The paths to the parent entities of this loop.
    """

    def __init__(
        self,
        index: int,
        workflow: Workflow,
        template: Loop,
        num_added_iterations: dict[tuple[int, ...], int],
        iterable_parameters: dict[str, IterableParam],
        output_parameters: dict[str, int],
        parents: list[str],
    ) -> None:
        self._index = index
        self._workflow = workflow
        self._template = template
        self._num_added_iterations = num_added_iterations
        self._iterable_parameters = iterable_parameters
        self._output_parameters = output_parameters
        self._parents = parents

        # appended to when adding an empty loop to the workflow that is a parent of this
        # loop; reset and added to `self._parents` on dump to disk:
        self._pending_parents: list[str] = []

        # used for `num_added_iterations` when a new loop iteration is added, or when
        # parents are append to; reset to None on dump to disk. Each key is a tuple of
        # parent loop indices and each value is the number of pending new iterations:
        self._pending_num_added_iterations: dict[tuple[int, ...], int] | None = None

        self._validate()

    @TimeIt.decorator
    def _validate(self) -> None:
        # task subset must be a contiguous range of task indices:
        task_indices = self.task_indices
        task_min, task_max = task_indices[0], task_indices[-1]
        if task_indices != tuple(range(task_min, task_max + 1)):
            raise LoopTaskSubsetError(self.name, self.task_indices)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(template={self.template!r}, "
            f"num_added_iterations={self.num_added_iterations!r})"
        )

    @property
    def num_added_iterations(self) -> Mapping[tuple[int, ...], int]:
        """
        The number of added iterations.
        """
        if self._pending_num_added_iterations:
            return self._pending_num_added_iterations
        else:
            return self._num_added_iterations

    @property
    def __pending(self) -> dict[tuple[int, ...], int]:
        if not self._pending_num_added_iterations:
            self._pending_num_added_iterations = dict(self._num_added_iterations)
        return self._pending_num_added_iterations

    def _initialise_pending_added_iters(self, added_iters: Iterable[int]):
        if not self._pending_num_added_iterations:
            self._pending_num_added_iterations = dict(self._num_added_iterations)
        if (added_iters_key := tuple(added_iters)) not in (pending := self.__pending):
            pending[added_iters_key] = 1

    def _increment_pending_added_iters(self, added_iters_key: Iterable[int]):
        self.__pending[tuple(added_iters_key)] += 1

    def _update_parents(self, parent: WorkflowLoop):
        assert parent.name
        self._pending_parents.append(parent.name)

        self._pending_num_added_iterations = {
            (*k, 0): v
            for k, v in (
                self._pending_num_added_iterations or self._num_added_iterations
            ).items()
        }

        self.workflow._store.update_loop_parents(
            index=self.index,
            num_added_iters=self.num_added_iterations,
            parents=self.parents,
        )

    def _reset_pending_num_added_iters(self) -> None:
        self._pending_num_added_iterations = None

    def _accept_pending_num_added_iters(self) -> None:
        if self._pending_num_added_iterations:
            self._num_added_iterations = dict(self._pending_num_added_iterations)
            self._reset_pending_num_added_iters()

    def _reset_pending_parents(self) -> None:
        self._pending_parents = []

    def _accept_pending_parents(self) -> None:
        self._parents += self._pending_parents
        self._reset_pending_parents()

    @property
    def index(self) -> int:
        """
        The index of this loop within its workflow.
        """
        return self._index

    @property
    def task_refs(self) -> tuple[int | str, ...]:
        """
        The list of task references (insert IDs or task unique names) that define the
        extent of the loop."""
        return self.template.task_refs

    @property
    def task_insert_IDs(self) -> tuple[int, ...]:
        """
        The insertion IDs of the tasks inside this loop.
        """
        return self.template.task_insert_IDs

    @property
    def task_objects(self) -> tuple[WorkflowTask, ...]:
        """
        The tasks in this loop.
        """
        return self.template.task_objects

    @property
    def task_indices(self) -> tuple[int, ...]:
        """
        The list of task indices that define the extent of the loop.
        """
        return tuple(task.index for task in self.task_objects)

    @property
    def workflow(self) -> Workflow:
        """
        The workflow containing this loop.
        """
        return self._workflow

    @property
    def template(self) -> Loop:
        """
        The loop template for this loop.
        """
        return self._template

    @property
    def parents(self) -> Sequence[str]:
        """
        The parents of this loop.
        """
        return self._parents + self._pending_parents

    @property
    def name(self) -> str:
        """
        The name of this loop, if one is defined.
        """
        assert self.template.name
        return self.template.name

    @property
    def iterable_parameters(self) -> dict[str, IterableParam]:
        """
        The parameters that are being iterated over.
        """
        return self._iterable_parameters

    @property
    def output_parameters(self) -> dict[str, int]:
        """
        The parameters that are outputs of this loop, and the final task insert ID from
        which each parameter is output.
        """
        return self._output_parameters

    @property
    def num_iterations(self) -> int:
        """
        The number of iterations.
        """
        return self.template.num_iterations

    @property
    def downstream_tasks(self) -> Iterator[WorkflowTask]:
        """Tasks that are not part of the loop, and downstream from this loop."""
        tasks = self.workflow.tasks
        for idx in range(self.task_objects[-1].index + 1, len(tasks)):
            yield tasks[idx]

    @property
    def upstream_tasks(self) -> Iterator[WorkflowTask]:
        """Tasks that are not part of the loop, and upstream from this loop."""
        tasks = self.workflow.tasks
        for idx in range(0, self.task_objects[0].index):
            yield tasks[idx]

    @staticmethod
    @TimeIt.decorator
    def _find_iterable_and_output_parameters(
        loop_template: Loop,
    ) -> tuple[dict[str, IterableParam], dict[str, int]]:
        all_inputs_first_idx: dict[str, int] = {}
        all_outputs_idx: dict[str, list[int]] = defaultdict(list)
        for task in loop_template.task_objects:
            for typ in task.template.all_schema_input_types:
                all_inputs_first_idx.setdefault(typ, task.insert_ID)
            for typ in task.template.all_schema_output_types:
                all_outputs_idx[typ].append(task.insert_ID)

        # find input parameters that are also output parameters at a later/same task:
        iterable_params: dict[str, IterableParam] = {}
        for typ, first_idx in all_inputs_first_idx.items():
            if typ in all_outputs_idx and first_idx <= all_outputs_idx[typ][0]:
                iterable_params[typ] = {
                    "input_task": first_idx,
                    "output_tasks": all_outputs_idx[typ],
                }

        for non_iter in loop_template.non_iterable_parameters:
            iterable_params.pop(non_iter, None)

        final_out_tasks = {k: v[-1] for k, v in all_outputs_idx.items()}

        return iterable_params, final_out_tasks

    @classmethod
    @TimeIt.decorator
    def new_empty_loop(
        cls,
        index: int,
        workflow: Workflow,
        template: Loop,
        iter_loop_idx: Sequence[Mapping[str, int]],
    ) -> WorkflowLoop:
        """
        Make a new empty loop.

        Parameters
        ----------
        index: int
            The index of the loop to create.
        workflow: ~hpcflow.app.Workflow
            The workflow that will contain the loop.
        template: Loop
            The template for the loop.
        iter_loop_idx: list[dict]
            Iteration information from parent loops.
        """
        parent_loops = cls._get_parent_loops(index, workflow, template)
        parent_names = [i.name for i in parent_loops]
        num_added_iters: dict[tuple[int, ...], int] = {}
        for i in iter_loop_idx:
            num_added_iters[tuple([i[j] for j in parent_names])] = 1

        iter_params, out_params = cls._find_iterable_and_output_parameters(template)
        return cls(
            index=index,
            workflow=workflow,
            template=template,
            num_added_iterations=num_added_iters,
            iterable_parameters=iter_params,
            output_parameters=out_params,
            parents=parent_names,
        )

    @classmethod
    @TimeIt.decorator
    def _get_parent_loops(
        cls,
        index: int,
        workflow: Workflow,
        template: Loop,
    ) -> list[WorkflowLoop]:
        parents: list[WorkflowLoop] = []
        passed_self = False
        self_tasks = set(template.task_insert_IDs)
        for loop_i in workflow.loops:
            if loop_i.index == index:
                passed_self = True
                continue
            other_tasks = set(loop_i.task_insert_IDs)
            if self_tasks.issubset(other_tasks):
                if (self_tasks == other_tasks) and not passed_self:
                    continue
                parents.append(loop_i)
        return parents

    @TimeIt.decorator
    def get_parent_loops(self) -> list[WorkflowLoop]:
        """Get loops whose task subset is a superset of this loop's task subset. If two
        loops have identical task subsets, the first loop in the workflow loop list is
        considered the child."""
        return self._get_parent_loops(self.index, self.workflow, self.template)

    @TimeIt.decorator
    def get_child_loops(self) -> list[WorkflowLoop]:
        """Get loops whose task subset is a subset of this loop's task subset. If two
        loops have identical task subsets, the first loop in the workflow loop list is
        considered the child."""
        children: list[WorkflowLoop] = []
        passed_self = False
        self_tasks = set(self.task_insert_IDs)
        for loop_i in self.workflow.loops:
            if loop_i.index == self.index:
                passed_self = True
                continue
            other_tasks = set(loop_i.task_insert_IDs)
            if self_tasks.issuperset(other_tasks):
                if (self_tasks == other_tasks) and passed_self:
                    continue
                children.append(loop_i)

        # order by depth, so direct child is first:
        return sorted(children, key=lambda x: len(next(iter(x.num_added_iterations))))

    @TimeIt.decorator
    def add_iteration(
        self,
        parent_loop_indices: Mapping[str, int] | None = None,
        cache: LoopCache | None = None,
        status: Status | None = None,
    ) -> None:
        """
        Add an iteration to this loop.

        Parameters
        ----------
        parent_loop_indices:
            Where have any parent loops got up to?
        cache:
            A cache used to make adding the iteration more efficient.
            One will be created if it is not supplied.
        """
        if not cache:
            cache = LoopCache.build(self.workflow)
        assert cache is not None
        parent_loops = self.get_parent_loops()
        child_loops = self.get_child_loops()
        parent_loop_indices_ = parent_loop_indices or {
            loop.name: 0 for loop in parent_loops
        }

        iters_key = tuple(parent_loop_indices_[p_nm] for p_nm in self.parents)
        cur_loop_idx = self.num_added_iterations[iters_key] - 1

        # keys are (task.insert_ID and element.index)
        all_new_data_idx: dict[tuple[int, int], DataIndex] = {}

        # initialise a new `num_added_iterations` key on each child loop:
        iters_key_dct = {
            **parent_loop_indices_,
            self.name: cur_loop_idx + 1,
        }
        for child in child_loops:
            child._initialise_pending_added_iters(
                iters_key_dct.get(j, 0) for j in child.parents
            )

            # needed for the case where an inner loop has only one iteration, meaning
            # `add_iteration` will not be called recursively on it:
            self.workflow._store.update_loop_num_iters(
                index=child.index,
                num_added_iters=child.num_added_iterations,
            )

        for task in self.task_objects:
            new_loop_idx = LoopIndex(iters_key_dct) + {
                child.name: 0
                for child in child_loops
                if task.insert_ID in child.task_insert_IDs
            }
            added_iter_IDs: list[int] = []
            for elem_idx in range(task.num_elements):
                elem_ID = task.element_IDs[elem_idx]

                new_data_idx: DataIndex = {}

                # copy resources from zeroth iteration:
                zeroth_iter_ID, zi_iter_data_idx = cache.zeroth_iters[elem_ID]
                zi_elem_ID, zi_idx = cache.iterations[zeroth_iter_ID]
                zi_data_idx = nth_value(cache.data_idx[zi_elem_ID], zi_idx)

                for key, val in zi_data_idx.items():
                    if key.startswith("resources."):
                        new_data_idx[key] = val

                for inp in task.template.all_schema_inputs:
                    is_inp_task = False
                    if iter_dat := self.iterable_parameters.get(inp.typ):
                        is_inp_task = task.insert_ID == iter_dat["input_task"]

                    inp_key = f"inputs.{inp.typ}"

                    if is_inp_task:
                        assert iter_dat is not None
                        inp_dat_idx = self.__get_looped_index(
                            task,
                            elem_ID,
                            cache,
                            iter_dat,
                            inp,
                            parent_loops,
                            parent_loop_indices_,
                            child_loops,
                            cur_loop_idx,
                        )
                        new_data_idx[inp_key] = inp_dat_idx
                    else:
                        orig_inp_src = cache.elements[elem_ID]["input_sources"][inp_key]
                        inp_dat_idx = None

                        if orig_inp_src.source_type is InputSourceType.LOCAL:
                            # keep locally defined inputs from original element
                            inp_dat_idx = zi_data_idx[inp_key]

                        elif orig_inp_src.source_type is InputSourceType.DEFAULT:
                            # keep default value from original element
                            try:
                                inp_dat_idx = zi_data_idx[inp_key]
                            except KeyError:
                                # if this input is required by a conditional action, and
                                # that condition is not met, then this input will not
                                # exist in the action-run data index, so use the initial
                                # iteration data index:
                                inp_dat_idx = zi_iter_data_idx[inp_key]

                        elif orig_inp_src.source_type is InputSourceType.TASK:
                            inp_dat_idx = self.__get_task_index(
                                task,
                                orig_inp_src,
                                cache,
                                elem_ID,
                                inp,
                                inp_key,
                                parent_loop_indices_,
                                all_new_data_idx,
                            )

                        if inp_dat_idx is None:
                            raise RuntimeError(
                                f"Could not find a source for parameter {inp.typ} "
                                f"when adding a new iteration for task {task!r}."
                            )

                        new_data_idx[inp_key] = inp_dat_idx

                # add any locally defined sub-parameters:
                inp_statuses = cache.elements[elem_ID]["input_statuses"]
                inp_status_inps = set(f"inputs.{inp}" for inp in inp_statuses)
                for sub_param_i in inp_status_inps.difference(new_data_idx):
                    sub_param_data_idx_iter_0 = zi_data_idx
                    try:
                        sub_param_data_idx = sub_param_data_idx_iter_0[sub_param_i]
                    except KeyError:
                        # as before, if this input is required by a conditional action,
                        # and that condition is not met, then this input will not exist in
                        # the action-run data index, so use the initial iteration data
                        # index:
                        sub_param_data_idx = zi_data_idx[sub_param_i]

                    new_data_idx[sub_param_i] = sub_param_data_idx

                for out in task.template.all_schema_outputs:
                    path_i = f"outputs.{out.typ}"
                    p_src: ParamSource = {"type": "EAR_output"}
                    new_data_idx[path_i] = self.workflow._add_unset_parameter_data(p_src)

                schema_params = set(i for i in new_data_idx if len(i.split(".")) == 2)
                all_new_data_idx[task.insert_ID, elem_idx] = new_data_idx

                iter_ID_i = self.workflow._store.add_element_iteration(
                    element_ID=elem_ID,
                    data_idx=new_data_idx,
                    schema_parameters=list(schema_params),
                    loop_idx=new_loop_idx,
                )
                if cache:
                    cache.add_iteration(
                        iter_ID=iter_ID_i,
                        task_insert_ID=task.insert_ID,
                        element_ID=elem_ID,
                        loop_idx=new_loop_idx,
                        data_idx=new_data_idx,
                    )

                added_iter_IDs.append(iter_ID_i)

            task.initialise_EARs(iter_IDs=added_iter_IDs)

        self._increment_pending_added_iters(
            parent_loop_indices_[p_nm] for p_nm in self.parents
        )
        self.workflow._store.update_loop_num_iters(
            index=self.index,
            num_added_iters=self.num_added_iterations,
        )

        # add iterations to fixed-number-iteration children only:
        for child in child_loops[::-1]:
            if child.num_iterations is not None:
                if status:
                    status_prev = str(status.status).rstrip(".")
                for iter_idx in range(child.num_iterations - 1):
                    if status:
                        status.update(
                            f"{status_prev} --> ({child.name!r}): iteration "
                            f"{iter_idx + 2}/{child.num_iterations}."
                        )
                    par_idx = {parent_name: 0 for parent_name in child.parents}
                    if parent_loop_indices:
                        par_idx.update(parent_loop_indices)
                    par_idx[self.name] = cur_loop_idx + 1
                    child.add_iteration(parent_loop_indices=par_idx, cache=cache)

        self.__update_loop_downstream_data_idx(parent_loop_indices_)

    def __get_src_ID_and_groups(
        self,
        elem_ID: int,
        iter_dat: IterableParam,
        inp: SchemaInput,
        cache: LoopCache,
        task: WorkflowTask,
    ) -> tuple[int, Sequence[int]]:
        # `cache.elements` contains only elements that are part of the
        # loop, so indexing a dependent element may raise:
        src_elem_IDs = {}
        for k, v in cache.element_dependents[elem_ID].items():
            try:
                if cache.elements[k]["task_insert_ID"] == iter_dat["output_tasks"][-1]:
                    src_elem_IDs[k] = v
            except KeyError:
                continue

        # consider groups
        single_data = inp.single_labelled_data
        assert single_data is not None
        inp_group_name = single_data.get("group")
        grouped_elems = [
            src_elem_j_ID
            for src_elem_j_ID, src_elem_j_dat in src_elem_IDs.items()
            if any(nm == inp_group_name for nm in src_elem_j_dat["group_names"])
        ]

        if not grouped_elems and len(src_elem_IDs) > 1:
            raise NotImplementedError(
                f"Multiple elements found in the iterable parameter "
                f"{inp!r}'s latest output task (insert ID: "
                f"{iter_dat['output_tasks'][-1]}) that can be used "
                f"to parametrise the next iteration of task "
                f"{task.unique_name!r}: "
                f"{list(src_elem_IDs)!r}."
            )

        elif not src_elem_IDs:
            # TODO: maybe OK?
            raise NotImplementedError(
                f"No elements found in the iterable parameter "
                f"{inp!r}'s latest output task (insert ID: "
                f"{iter_dat['output_tasks'][-1]}) that can be used "
                f"to parametrise the next iteration."
            )

        return nth_key(src_elem_IDs, 0), grouped_elems

    def __get_looped_index(
        self,
        task: WorkflowTask,
        elem_ID: int,
        cache: LoopCache,
        iter_dat: IterableParam,
        inp: SchemaInput,
        parent_loops: list[WorkflowLoop],
        parent_loop_indices: Mapping[str, int],
        child_loops: list[WorkflowLoop],
        cur_loop_idx: int,
    ):
        # source from final output task of previous iteration, with all parent
        # loop indices the same as previous iteration, and all child loop indices
        # maximised:

        # identify element(s) from which this iterable input should be
        # parametrised:
        if task.insert_ID == iter_dat["output_tasks"][-1]:
            # single-task loop
            src_elem_ID = elem_ID
            grouped_elems: Sequence[int] = []
        else:
            # multi-task loop
            src_elem_ID, grouped_elems = self.__get_src_ID_and_groups(
                elem_ID, iter_dat, inp, cache, task
            )

        child_loop_max_iters: dict[str, int] = {}
        parent_loop_same_iters = {
            loop.name: parent_loop_indices[loop.name] for loop in parent_loops
        }
        child_iter_parents = {
            **parent_loop_same_iters,
            self.name: cur_loop_idx,
        }
        for loop in child_loops:
            if iter_dat["output_tasks"][-1] in loop.task_insert_IDs:
                i_num_iters = loop.num_added_iterations[
                    tuple(child_iter_parents[j] for j in loop.parents)
                ]
                i_max = i_num_iters - 1
                child_iter_parents[loop.name] = i_max
                child_loop_max_iters[loop.name] = i_max

        loop_idx_key = LoopIndex(child_loop_max_iters)
        loop_idx_key.update(parent_loop_same_iters)
        loop_idx_key[self.name] = cur_loop_idx

        # identify the ElementIteration from which this input should be
        # parametrised:
        if grouped_elems:
            src_data_idx = [
                cache.data_idx[src_elem_ID][loop_idx_key] for src_elem_ID in grouped_elems
            ]
            if not src_data_idx:
                raise RuntimeError(
                    f"Could not find a source iteration with loop_idx: "
                    f"{loop_idx_key!r}."
                )
            return [i[f"outputs.{inp.typ}"] for i in src_data_idx]
        else:
            return cache.data_idx[src_elem_ID][loop_idx_key][f"outputs.{inp.typ}"]

    def __get_task_index(
        self,
        task: WorkflowTask,
        orig_inp_src: InputSource,
        cache: LoopCache,
        elem_ID: int,
        inp: SchemaInput,
        inp_key: str,
        parent_loop_indices: Mapping[str, int],
        all_new_data_idx: Mapping[tuple[int, int], DataIndex],
    ) -> int | list[int]:
        if orig_inp_src.task_ref not in self.task_insert_IDs:
            # source the data_idx from the iteration with same parent
            # loop indices as the new iteration to add:
            src_data_idx = next(
                di_k
                for li_k, di_k in cache.data_idx[elem_ID].items()
                if all(li_k.get(p_k) == p_v for p_k, p_v in parent_loop_indices.items())
            )

            # could be multiple, but they should all have the same
            # data index for this parameter:
            return src_data_idx[inp_key]

        is_group = (
            inp.single_labelled_data is not None
            and "group" in inp.single_labelled_data
            # this input is a group, assume for now all elements
        )

        # same task/element, but update iteration to the just-added
        # iteration:
        assert orig_inp_src.task_source_type is not None
        key_prefix = orig_inp_src.task_source_type.name.lower()
        prev_dat_idx_key = f"{key_prefix}s.{inp.typ}"
        new_sources: list[tuple[int, int]] = []
        for (tiID, e_idx), _ in all_new_data_idx.items():
            if tiID == orig_inp_src.task_ref:
                # find which element in that task `element`
                # depends on:
                src_elem_IDs = cache.element_dependents[
                    self.workflow.tasks.get(insert_ID=tiID).element_IDs[e_idx]
                ]
                # `cache.elements` contains only elements that are part of the loop, so
                # indexing a dependent element may raise:
                src_elem_IDs_i = []
                for k, _v in src_elem_IDs.items():
                    try:
                        if (
                            cache.elements[k]["task_insert_ID"] == task.insert_ID
                            and k == elem_ID
                            # filter src_elem_IDs_i for matching element IDs
                        ):

                            src_elem_IDs_i.append(k)
                    except KeyError:
                        continue

                if len(src_elem_IDs_i) == 1:
                    new_sources.append((tiID, e_idx))

        if is_group:
            # Convert into simple list of indices
            return list(
                chain.from_iterable(
                    self.__as_sequence(all_new_data_idx[src][prev_dat_idx_key])
                    for src in new_sources
                )
            )
        else:
            assert len(new_sources) == 1
            return all_new_data_idx[new_sources[0]][prev_dat_idx_key]

    @staticmethod
    def __as_sequence(seq: int | Iterable[int]) -> Iterable[int]:
        if isinstance(seq, int):
            yield seq
        else:
            yield from seq

    def __update_loop_downstream_data_idx(
        self,
        parent_loop_indices: Mapping[str, int],
    ):
        # update data indices of loop-downstream tasks that depend on task outputs from
        # this loop:

        # keys: iter or run ID, values: dict of param type and new parameter index
        iter_new_data_idx: dict[int, DataIndex] = defaultdict(dict)
        run_new_data_idx: dict[int, DataIndex] = defaultdict(dict)

        param_sources = self.workflow.get_all_parameter_sources()

        # keys are parameter type, then task insert ID, then data index keys mapping to
        # their updated values:
        all_updates: dict[str, dict[int, dict[int, int]]] = defaultdict(
            lambda: defaultdict(dict)
        )

        for task in self.downstream_tasks:
            for elem in task.elements:
                for param_typ, param_out_task_iID in self.output_parameters.items():
                    if param_typ in task.template.all_schema_input_types:
                        # this element's input *might* need updating, only if it has a
                        # task input source type that is this loop's output task for this
                        # parameter:
                        elem_src = elem.input_sources[f"inputs.{param_typ}"]
                        if (
                            elem_src.source_type is InputSourceType.TASK
                            and elem_src.task_source_type is TaskSourceType.OUTPUT
                            and elem_src.task_ref == param_out_task_iID
                        ):
                            for iter_i in elem.iterations:

                                # do not modify element-iterations of previous iterations
                                # of the current loop:
                                skip_iter = False
                                for k, v in parent_loop_indices.items():
                                    if iter_i.loop_idx.get(k) != v:
                                        skip_iter = True
                                        break

                                if skip_iter:
                                    continue

                                # update the iteration data index and any pending runs:
                                iter_old_di = iter_i.data_idx[f"inputs.{param_typ}"]

                                is_group = True
                                if not isinstance(iter_old_di, list):
                                    is_group = False
                                    iter_old_di = [iter_old_di]

                                iter_old_run_source = [
                                    param_sources[i]["EAR_ID"] for i in iter_old_di
                                ]
                                iter_old_run_objs = self.workflow.get_EARs_from_IDs(
                                    iter_old_run_source
                                )  # TODO: use cache

                                # need to check the run source is actually from the loop
                                # output task (it could be from a previous iteration of a
                                # separate loop in this task):
                                if any(
                                    i.task.insert_ID != param_out_task_iID
                                    for i in iter_old_run_objs
                                ):
                                    continue

                                iter_new_iters = [
                                    i.element.iterations[-1] for i in iter_old_run_objs
                                ]

                                # note: we can cast to int, because output keys never
                                # have multiple data indices (unlike input keys):
                                iter_new_dis = [
                                    cast("int", i.get_data_idx()[f"outputs.{param_typ}"])
                                    for i in iter_new_iters
                                ]

                                # keep track of updates so we can also update task-input
                                # type sources:
                                all_updates[param_typ][task.insert_ID].update(
                                    dict(zip(iter_old_di, iter_new_dis))
                                )

                                iter_new_data_idx[iter_i.id_][f"inputs.{param_typ}"] = (
                                    iter_new_dis if is_group else iter_new_dis[0]
                                )

                                for run_j in iter_i.action_runs:
                                    if run_j.status is EARStatus.pending:
                                        try:
                                            old_di = run_j.data_idx[f"inputs.{param_typ}"]
                                        except KeyError:
                                            # not all actions will include this input
                                            continue

                                        is_group = True
                                        if not isinstance(old_di, list):
                                            is_group = False
                                            old_di = [old_di]

                                        old_run_source = [
                                            param_sources[i]["EAR_ID"] for i in old_di
                                        ]
                                        old_run_objs = self.workflow.get_EARs_from_IDs(
                                            old_run_source
                                        )  # TODO: use cache

                                        # need to check the run source is actually from the loop
                                        # output task (it could be from a previous action in this
                                        # element-iteration):
                                        if any(
                                            i.task.insert_ID != param_out_task_iID
                                            for i in old_run_objs
                                        ):
                                            continue

                                        new_iters = [
                                            i.element.iterations[-1] for i in old_run_objs
                                        ]

                                        # note: we can cast to int, because output keys
                                        # never have multiple data indices (unlike input
                                        # keys):
                                        new_dis = [
                                            cast(
                                                "int",
                                                i.get_data_idx()[f"outputs.{param_typ}"],
                                            )
                                            for i in new_iters
                                        ]

                                        run_new_data_idx[run_j.id_][
                                            f"inputs.{param_typ}"
                                        ] = (new_dis if is_group else new_dis[0])

                        elif (
                            elem_src.source_type is InputSourceType.TASK
                            and elem_src.task_source_type is TaskSourceType.INPUT
                        ):
                            # parameters are that sourced from inputs of other tasks,
                            # might need to be updated if those other tasks have
                            # themselves had their data indices updated:
                            assert elem_src.task_ref
                            ups_i = all_updates.get(param_typ, {}).get(elem_src.task_ref)

                            if ups_i:
                                # if a further-downstream task has a task-input source
                                # that points to this task, this will also need updating:
                                all_updates[param_typ][task.insert_ID].update(ups_i)

                            else:
                                continue

                            for iter_i in elem.iterations:

                                # update the iteration data index and any pending runs:
                                iter_old_di = iter_i.data_idx[f"inputs.{param_typ}"]

                                is_group = True
                                if not isinstance(iter_old_di, list):
                                    is_group = False
                                    iter_old_di = [iter_old_di]

                                iter_new_dis = [ups_i.get(i, i) for i in iter_old_di]

                                if iter_new_dis != iter_old_di:
                                    iter_new_data_idx[iter_i.id_][
                                        f"inputs.{param_typ}"
                                    ] = (iter_new_dis if is_group else iter_new_dis[0])

                                for run_j in iter_i.action_runs:
                                    if run_j.status is EARStatus.pending:
                                        try:
                                            old_di = run_j.data_idx[f"inputs.{param_typ}"]
                                        except KeyError:
                                            # not all actions will include this input
                                            continue

                                        is_group = True
                                        if not isinstance(old_di, list):
                                            is_group = False
                                            old_di = [old_di]

                                        new_dis = [ups_i.get(i, i) for i in old_di]

                                        if new_dis != old_di:
                                            run_new_data_idx[run_j.id_][
                                                f"inputs.{param_typ}"
                                            ] = (new_dis if is_group else new_dis[0])

        # now update data indices (TODO: including in cache!)
        if iter_new_data_idx:
            self.workflow._store.update_iter_data_indices(iter_new_data_idx)

        if run_new_data_idx:
            self.workflow._store.update_run_data_indices(run_new_data_idx)

    def test_termination(self, element_iter) -> bool:
        """Check if a loop should terminate, given the specified completed element
        iteration."""
        if self.template.termination:
            return self.template.termination.test(element_iter)
        return False

    @TimeIt.decorator
    def get_element_IDs(self):
        elem_IDs = [
            j
            for i in self.task_insert_IDs
            for j in self.workflow.tasks.get(insert_ID=i).element_IDs
        ]
        return elem_IDs

    @TimeIt.decorator
    def get_elements(self):
        return self.workflow.get_elements_from_IDs(self.get_element_IDs())

    @TimeIt.decorator
    def skip_downstream_iterations(self, elem_iter) -> list[int]:
        """
        Parameters
        ----------
        elem_iter
            The element iteration whose subsequent iterations should be skipped.
        dep_element_IDs
            List of elements that are dependent (recursively) on the element
            of `elem_iter`.
        """
        current_iter_idx = elem_iter.loop_idx[self.name]
        current_task_iID = elem_iter.task.insert_ID
        self._app.logger.info(
            f"setting loop {self.name!r} iterations downstream of current iteration "
            f"index {current_iter_idx} to skip"
        )
        elements = self.get_elements()

        # TODO: fix for multiple loop cycles
        warn(
            "skip downstream iterations does not work correctly for multiple loop cycles!"
        )

        to_skip = []
        for elem in elements:
            for iter_i in elem.iterations:
                if iter_i.loop_idx[self.name] > current_iter_idx or (
                    iter_i.loop_idx[self.name] == current_iter_idx
                    and iter_i.task.insert_ID > current_task_iID
                ):
                    to_skip.extend(iter_i.EAR_IDs_flat)

        self._app.logger.info(
            f"{len(to_skip)} runs will be set to skip: {shorten_list_str(to_skip)}"
        )
        self.workflow.set_EAR_skip({k: SkipReason.LOOP_TERMINATION for k in to_skip})

        return to_skip



"""
Helper for running a subprocess.
"""

from __future__ import annotations
import subprocess
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from logging import Logger


def run_cmd(
    cmd: str | Sequence[str], logger: Logger | None = None, **kwargs
) -> tuple[str, str]:
    """Execute a command and return stdout, stderr as strings."""
    if logger:
        logger.debug(f"running shell command: {cmd}")
    proc = subprocess.run(
        args=cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs
    )
    stdout = proc.stdout.decode()
    stderr = proc.stderr.decode()
    return stdout, stderr



"""
Types used in configuration.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import TypeAlias, TypedDict, TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from pathlib import Path
    from typing import Any
    from typing_extensions import NotRequired
    from .config import Config
    from ..core.validation import Schema


T = TypeVar("T")
#: Type of a getter callback.
GetterCallback: TypeAlias = "Callable[[Config, T], T]"
#: Type of a setter callback.
SetterCallback: TypeAlias = "Callable[[Config, T], Any]"
#: Type of a unsetter callback.
UnsetterCallback: TypeAlias = "Callable[[Config], None]"


class SGEParallelEnvsDescriptor(TypedDict):
    """
    SGE parallel environment descriptor.
    """

    #: Number of supported cores.
    num_cores: list[int]


class SLURMPartitionsDescriptor(TypedDict):
    """
    SLURM partition descriptor.
    """

    #: Number of supported cores.
    num_cores: NotRequired[list[int]]
    #: Number of cores per node.
    num_cores_per_node: NotRequired[list[int]]
    #: Number of supported nodes.
    num_nodes: NotRequired[list[int]]
    #: Possible parallel modes.
    parallel_modes: NotRequired[list[str]]


class SchedulerConfigDescriptor(TypedDict):
    """
    Scheduler configuration descriptor.
    """

    #: Default values.
    defaults: dict[str, Any]
    #: SGE parallel environments, by name.
    parallel_environments: NotRequired[dict[str, SGEParallelEnvsDescriptor]]
    #: SLURM partitions, by name.
    partitions: NotRequired[dict[str, SLURMPartitionsDescriptor]]


class ShellConfigDescriptor(TypedDict):
    """
    Shell configuration descriptor.
    """

    #: Default values.
    defaults: dict[str, Any]


class ConfigDescriptor(TypedDict):
    """
    Configuration descriptor.
    """

    #: Execution machine.
    machine: NotRequired[str]
    #: Where to log.
    log_file_path: NotRequired[str]
    #: Where to find execution environments.
    environment_sources: NotRequired[list[str]]
    #: Where to find task schemas.
    task_schema_sources: NotRequired[list[str]]
    #: Where to find command files.
    command_file_sources: NotRequired[list[str]]
    #: Where to find parameter implementations.
    parameter_sources: NotRequired[list[str]]
    #: Default scheduler.
    default_scheduler: NotRequired[str]
    #: Default shell.
    default_shell: NotRequired[str]
    #: Supported schedulers.
    schedulers: NotRequired[dict[str, SchedulerConfigDescriptor]]
    #: Supported shells.
    shells: NotRequired[dict[str, ShellConfigDescriptor]]
    #: User affiliations
    user_affiliations: NotRequired[list[str]]


class InvocationDescriptor(TypedDict):
    """
    Invocation descriptor.
    """

    #: Used to set up the environment.
    environment_setup: str | None
    #: setting to apply if matched.
    match: dict[str, str | list[str]]


class DefaultConfiguration(TypedDict):
    """
    The default configuration.
    """

    #: Default invocation.
    invocation: InvocationDescriptor
    #: Default configuration.
    config: ConfigDescriptor


#: A configuration dictionary.
ConfigDict: TypeAlias = "dict[str, dict[str, DefaultConfiguration]]"


class ConfigMetadata(TypedDict):
    """
    Metadata supported by the :class:`Config` class.
    """

    #: Location of directory containing the config file.
    config_directory: Path
    #: Name of the config file.
    config_file_name: str
    #: Full path to the config file.
    config_file_path: Path
    #: The contents of the config file.
    config_file_contents: str
    #: The key identifying the config section within the config file.
    config_key: str
    #: Schemas that apply to the config.
    config_schemas: Sequence[Schema]
    #: The user that invoked things.
    invoking_user_id: str
    #: The user hosting things.
    host_user_id: str
    #: Path to file holding description of :attr:``host_user_id``.
    host_user_id_file_path: Path



def main_script_test_direct_in_direct_out_2_fail_allowed_group(p2):
    # process, accounting for possible unset data:
    p2_sum = sum(i for i in p2 if i is not None)
    p3 = p2_sum + 100

    # return outputs
    return {"p3": p3}



"""
Schema management.
"""

from __future__ import annotations
from collections.abc import Mapping, Sequence
from typing import Any, Generic, Protocol, TypeVar
from valida import Schema as ValidaSchema  # type: ignore
from hpcflow.sdk.core.utils import open_text_resource

T = TypeVar("T")


class ValidatedData(Protocol, Generic[T]):
    """
    Typed profile of ``valida.ValidatedData``.
    """

    @property
    def is_valid(self) -> bool: ...

    def get_failures_string(self) -> str: ...

    cast_data: T


class PreparedConditionCallable(Protocol):
    """
    Typed profile of ``valida.PreparedConditionCallable``.
    """

    @property
    def name(self) -> str: ...

    @property
    def args(self) -> tuple[str, ...]: ...


class Condition(Protocol):
    """
    Typed profile of ``valida.Condition``.
    """

    @property
    def callable(self) -> PreparedConditionCallable: ...


class Rule(Protocol):
    """
    Typed profile of ``valida.Rule``.
    """

    @property
    def condition(self) -> Condition: ...

    @property
    def path(self) -> object: ...


class Schema(Protocol):
    """
    Typed profile of ``valida.Schema``.
    """

    def validate(self, data: T) -> ValidatedData[T]: ...

    @property
    def rules(self) -> Sequence[Rule]: ...

    def add_schema(self, schema: Schema, root_path: Any = None) -> None: ...

    def to_tree(self, **kwargs) -> Sequence[Mapping[str, str]]: ...


def get_schema(filename) -> Schema:
    """
    Get a valida `Schema` object from the embedded data directory.

    Parameter
    ---------
    schema: str
        The name of the schema file within the resources package
        (:py:mod:`hpcflow.sdk.data`).
    """
    with open_text_resource("hpcflow.sdk.data", filename) as fh:
        schema_dat = fh.read()
    return ValidaSchema.from_yaml(schema_dat)



from pathlib import Path
from hpcflow.sdk.core.cache import ObjectCache
from hpcflow.sdk.core.test_utils import make_workflow


def test_object_cache_dependencies_simple(tmp_path: Path):
    wk = make_workflow(
        schemas_spec=[
            ({"p1": None}, ("p2",), "t1"),
            ({"p2": None}, ("p3",), "t2"),
            ({"p3": None}, ("p4",), "t3"),
            ({"p4": None}, ("p5",), "t4"),
        ],
        path=tmp_path,
        local_inputs={0: ("p1",)},
        overwrite=True,
    )
    obj_cache = ObjectCache.build(wk, dependencies=True)
    assert obj_cache.run_dependencies == {0: set(), 1: {0}, 2: {1}, 3: {2}}
    assert obj_cache.run_dependents == {0: {1}, 1: {2}, 2: {3}, 3: set()}
    assert obj_cache.iter_run_dependencies == {0: set(), 1: {0}, 2: {1}, 3: {2}}
    assert obj_cache.iter_iter_dependencies == {
        0: set(),
        1: {0},
        2: {1},
        3: {2},
    }
    assert obj_cache.elem_iter_dependencies == {
        0: set(),
        1: {0},
        2: {1},
        3: {2},
    }
    assert obj_cache.elem_elem_dependencies == {
        0: set(),
        1: {0},
        2: {1},
        3: {2},
    }
    assert obj_cache.elem_elem_dependents == {0: {1}, 1: {2}, 2: {3}, 3: set()}
    assert obj_cache.elem_elem_dependents_rec == {
        0: {1, 2, 3},
        1: {2, 3},
        2: {3},
        3: set(),
    }



"""
Model of an execution environment.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.core.errors import DuplicateExecutableError
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.object_list import ExecutablesList
from hpcflow.sdk.core.utils import check_valid_py_identifier, get_duplicate_items

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, ClassVar


@dataclass
class NumCores(JSONLike):
    """
    A range of cores supported by an executable instance.

    Parameters
    ----------
    start:
        The minimum number of cores supported.
    stop:
        The maximum number of cores supported.
    step: int
        The step in the number of cores supported. Defaults to 1.
    """

    #: The minimum number of cores supported.
    start: int
    #: The maximum number of cores supported.
    stop: int
    #: The step in the number of cores supported. Normally 1.
    step: int = 1

    def __contains__(self, x: int) -> bool:
        return x in range(self.start, self.stop + 1, self.step)


@dataclass
@hydrate
class ExecutableInstance(JSONLike):
    """
    A particular instance of an executable that can support some mode of operation.

    Parameters
    ----------
    parallel_mode:
        What parallel mode is supported by this executable instance.
    num_cores: NumCores | int | dict[str, int]
        The number of cores supported by this executable instance.
    command:
        The actual command to use for this executable instance.
    """

    #: What parallel mode is supported by this executable instance.
    parallel_mode: str | None
    #: The number of cores supported by this executable instance.
    num_cores: NumCores
    #: The actual command to use for this executable instance.
    command: str

    def __init__(
        self, parallel_mode: str | None, num_cores: NumCores | int | dict, command: str
    ):
        self.parallel_mode = parallel_mode
        self.command = command
        if isinstance(num_cores, NumCores):
            self.num_cores = num_cores
        elif isinstance(num_cores, int):
            self.num_cores = NumCores(num_cores, num_cores)
        else:
            self.num_cores = NumCores(**num_cores)

    @classmethod
    def from_spec(cls, spec: dict[str, Any]) -> ExecutableInstance:
        """
        Construct an instance from a specification dictionary.
        """
        return cls(**spec)


class Executable(JSONLike):
    """
    A program managed by the environment.

    Parameters
    ----------
    label:
        The abstract name of the program.
    instances: list[ExecutableInstance]
        The concrete instances of the application that may be present.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="instances",
            class_name="ExecutableInstance",
            is_multiple=True,
        ),
    )

    def __init__(self, label: str, instances: list[ExecutableInstance]):
        #: The abstract name of the program.
        self.label = check_valid_py_identifier(label)
        #: The concrete instances of the application that may be present.
        self.instances = instances

        self._executables_list: ExecutablesList | None = None  # assigned by parent

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"label={self.label}, "
            f"instances={self.instances!r}"
            f")"
        )

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and self.label == other.label
            and self.instances == other.instances
            and (
                (
                    self.environment
                    and other.environment
                    and self.environment.name == other.environment.name
                )
                or (not self.environment and not other.environment)
            )
        )

    @property
    def environment(self) -> Environment | None:
        """
        The environment that the executable is going to run in.
        """
        return None if (el := self._executables_list) is None else el.environment

    def filter_instances(
        self, parallel_mode: str | None = None, num_cores: int | None = None
    ) -> list[ExecutableInstance]:
        """
        Select the instances of the executable that are compatible with the given
        requirements.

        Parameters
        ----------
        parallel_mode: str
            If given, the parallel mode to require.
        num_cores:  int
            If given, the number of cores desired.

        Returns
        -------
        list[ExecutableInstance]:
            The known executable instances that match the requirements.
        """
        return [
            inst
            for inst in self.instances
            if (parallel_mode is None or inst.parallel_mode == parallel_mode)
            and (num_cores is None or num_cores in inst.num_cores)
        ]


class Environment(JSONLike):
    """
    An execution environment that contains a number of executables.

    Parameters
    ----------
    name: str
        The name of the environment.
    setup: list[str]
        Commands to run to enter the environment.
    specifiers: dict[str, str]
        Dictionary of attributes that may be used to supply additional key/value pairs to
        look up an environment by.
    executables: list[Executable]
        List of abstract executables in the environment.
    """

    _validation_schema: ClassVar[str] = "environments_spec_schema.yaml"
    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="executables",
            class_name="ExecutablesList",
            parent_ref="environment",
        ),
    )

    def __init__(
        self,
        name: str,
        setup: Sequence[str] | None = None,
        specifiers: Mapping[str, str] | None = None,
        executables: ExecutablesList | Sequence[Executable] | None = None,
        doc: str = "",
        _hash_value: str | None = None,
    ):
        #: The name of the environment.
        self.name = name
        #: Documentation for the environment.
        self.doc = doc
        #: Dictionary of attributes that may be used to supply additional key/value pairs
        #: to look up an environment by.
        self.specifiers: Mapping[str, str] = specifiers or {}
        #: List of abstract executables in the environment.
        self.executables = (
            executables
            if isinstance(executables, ExecutablesList)
            else self._app.ExecutablesList(executables or ())
        )
        self._hash_value = _hash_value
        #: Commands to run to enter the environment.
        self.setup: tuple[str, ...] | None
        if not setup:
            self.setup = None
        elif isinstance(setup, str):
            self.setup = tuple(cmd.strip() for cmd in setup.strip().split("\n"))
        else:
            self.setup = tuple(setup)
        self._set_parent_refs()
        self._validate()

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and self.setup == other.setup
            and self.executables == other.executables
            and self.specifiers == other.specifiers
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def _validate(self):
        if dup_labels := get_duplicate_items(exe.label for exe in self.executables):
            raise DuplicateExecutableError(dup_labels)

    @property
    def documentation(self) -> str:
        if self.doc:
            import markupsafe

            return markupsafe.Markup(self.doc)
        return repr(self)



from __future__ import annotations


def import_future_script():
    """Used to test the `__future__` import is moved to the top of the file in the
    generated script in the case of `resources.combine_scripts == True`"""
    pass



from typing import Iterable
import re


def shorten_list_str(
    lst: Iterable, items: int = 10, end_num: int = 1, placeholder: str = "..."
) -> str:
    """Format a list as a string, including only some maximum number of items.

    Parameters
    ----------
    lst:
        The list to format in a shortened form.
    items:
        The total number of items to include in the formatted list.
    end_num:
        The number of items to include at the end of the formatted list.
    placeholder
        The placeholder to use to replace excess items in the formatted list.

    Examples
    --------
    >>> shorten_list_str(list(range(20)), items=5)
    '[0, 1, 2, 3, ..., 19]'

    """
    lst = list(lst)
    if len(lst) <= items + 1:  # (don't replace only one item)
        lst_short = lst
    else:
        start_num = items - end_num
        lst_short = lst[:start_num] + ["..."] + lst[-end_num:]

    return "[" + ", ".join(f"{i}" for i in lst_short) + "]"


def extract_py_from_future_imports(py_str: str) -> tuple[str, set[str]]:
    """
    Remove any `from __future__ import <feature>` lines from a string of Python code, and
    return the modified string, and a list of `<feature>`s that were imported.

    Notes
    -----
    This is required when generated a combined-scripts jobscript that concatenates
    multiple Python scripts into one script. If `__future__` statements are included in
    these individual scripts, they must be moved to the top of the file [1].

    References
    ----------
    [1] https://docs.python.org/3/reference/simple_stmts.html#future-statements

    """

    pattern = r"^from __future__ import (.*)\n"
    if future_imports := (set(re.findall(pattern, py_str, flags=re.MULTILINE) or ())):
        future_imports = {
            j.strip() for i in future_imports for j in i.split(",") if j.strip()
        }
        py_str = re.sub(pattern, "", py_str, flags=re.MULTILINE)

    return (py_str, future_imports)


def capitalise_first_letter(chars: str) -> str:
    """
    Convert the first character of a string to upper case (if that makes sense).
    The rest of the string is unchanged.
    """
    return chars[0].upper() + chars[1:]



from __future__ import annotations
import os
import time
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.config.errors import (
    ConfigFileValidationError,
    ConfigItemCallbackError,
    ConfigNonConfigurableError,
    ConfigReadOnlyError,
)


def test_reset_config(new_null_config) -> None:
    cfg_dir = hf.config.config_directory
    machine_name = hf.config.machine
    new_machine_name = machine_name + "123"
    hf.config.machine = new_machine_name
    assert hf.config.machine == new_machine_name
    hf.reset_config(config_dir=cfg_dir)
    assert hf.config.machine == machine_name


def test_raise_on_invalid_config_file(new_null_config) -> None:
    # make an invalid config file:
    cfg_path = hf.config.config_file_path
    with cfg_path.open("at+") as f:
        f.write("something_invalid: 1\n")

    # try to load the invalid file:
    cfg_dir = hf.config.config_directory
    with pytest.raises(ConfigFileValidationError):
        hf.reload_config(config_dir=cfg_dir, warn=False)
    hf.reset_config(config_dir=cfg_dir, warn=False)
    hf.unload_config()


def test_reset_invalid_config(new_null_config) -> None:
    # make an invalid config file:
    cfg_path = hf.config.config_file_path
    with cfg_path.open("at+") as f:
        f.write("something_invalid: 1\n")

    # check we can reset the invalid file:
    cfg_dir = hf.config.config_directory
    hf.reset_config(config_dir=cfg_dir, warn=False)
    hf.unload_config()


def test_raise_on_set_default_scheduler_not_in_schedulers_list_invalid_name(
    null_config,
) -> None:
    new_default = "invalid-scheduler"
    with pytest.raises(ConfigItemCallbackError):
        hf.config.default_scheduler = new_default


def test_raise_on_set_default_scheduler_not_in_schedulers_list_valid_name(
    null_config,
) -> None:
    new_default = "slurm"  # valid but unsupported (by default) scheduler
    with pytest.raises(ConfigItemCallbackError):
        hf.config.default_scheduler = new_default


def test_without_callbacks_ctx_manager(null_config) -> None:
    # set a new shell that would raise an error in the `callback_supported_shells`:
    new_default = "bash" if os.name == "nt" else "powershell"

    with hf.config._without_callbacks("callback_supported_shells"):
        hf.config.default_shell = new_default
        assert hf.config.default_shell == new_default

    # outside the context manager, the callback is reinstated, which should raise:
    with pytest.raises(ConfigItemCallbackError):
        hf.config.default_shell

    # unload the modified config so it's not reused by other tests
    hf.unload_config()


@pytest.mark.xfail(reason="Might occasionally fail.")
def test_cache_faster_than_no_cache(null_config):
    n = 10_000
    tic = time.perf_counter()
    for _ in range(n):
        _ = hf.config.machine
    toc = time.perf_counter()
    elapsed_no_cache = toc - tic

    with hf.config.cached_config():
        tic = time.perf_counter()
        for _ in range(n):
            _ = hf.config.machine
        toc = time.perf_counter()
    elapsed_cache = toc - tic

    assert elapsed_cache < elapsed_no_cache


def test_cache_read_only(new_null_config):
    """Check we cannot modify the config when using the cache"""

    # check we can set an item first:
    hf.machine = "abc"
    assert hf.machine == "abc"

    with pytest.raises(ConfigReadOnlyError):
        with hf.config.cached_config():
            hf.config.set("machine", "123")

    with pytest.raises(ConfigReadOnlyError):
        with hf.config.cached_config():
            hf.config.machine = "456"


def test_workflow_template_config_validation(new_null_config, tmp_path):
    wkt = hf.WorkflowTemplate(
        tasks=[],
        config={"log_file_level": "debug"},
        name="test_workflow_config_validation",
    )
    assert wkt.config == {"log_file_level": "debug"}


def test_workflow_template_config_validation_raises(unload_config, tmp_path):
    with pytest.raises(ConfigNonConfigurableError):
        hf.WorkflowTemplate(
            tasks=[],
            config={"bad_key": "debug"},
            name="test_workflow_config_validation_raises",
        )

    # workflow template config validation should not need to load the whole config:
    assert not hf.is_config_loaded


def test_config_with_updates(new_null_config):
    level_1 = hf.config.get("log_console_level")
    with hf.config._with_updates({"log_console_level": "debug"}):
        level_2 = hf.config.get("log_console_level")
    level_3 = hf.config.get("log_console_level")
    assert level_1 == level_3 != level_2
    hf.reload_config()


@pytest.mark.integration
def test_workflow_template_config_set(new_null_config, tmp_path):
    """Test we can set a workflow-level config item and that it is correctly applied
    during execution."""

    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 101},
    )
    log_path = tmp_path / "log.log"
    hf.config.set("log_file_level", "warning")
    hf.config.set("log_file_path", log_path)

    log_str_1 = "this should not appear in the log file"
    hf.submission_logger.debug(log_str_1)

    log_str_2 = "this should appear in the log file"
    hf.submission_logger.warning(log_str_2)

    assert log_path.is_file()
    log_file_contents = log_path.read_text()
    assert log_str_1 not in log_file_contents
    assert log_str_2 in log_file_contents

    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        config={"log_file_level": "debug"},
        resources={"any": {"write_app_logs": True}},
        workflow_name="test_workflow_config",
        template_name="test_workflow_config",
        path=tmp_path,
    )
    wk.submit(wait=True, status=False, add_to_known=False)

    # check some DEBUG messages present in the run logs
    debug_str = " DEBUG hpcflow.persistence:"

    run = wk.get_EARs_from_IDs([0])[0]
    run_log_path = run.get_app_log_path()
    assert run_log_path.is_file()

    run_log_contents = run_log_path.read_text()
    assert debug_str in run_log_contents

    # log file level should not have changed:
    assert hf.config.get("log_file_level") == "warning"

    hf.reload_config()



"""
An interface to SLURM.
"""

from __future__ import annotations
import subprocess
import time
from typing import cast, TYPE_CHECKING
from typing_extensions import override
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.core.enums import ParallelMode
from hpcflow.sdk.core.errors import (
    IncompatibleParallelModeError,
    IncompatibleSLURMArgumentsError,
    IncompatibleSLURMPartitionError,
    UnknownSLURMPartitionError,
)
from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.submission.enums import JobscriptElementState
from hpcflow.sdk.submission.schedulers import QueuedScheduler
from hpcflow.sdk.submission.schedulers.utils import run_cmd

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Iterator, Mapping, Sequence
    from typing import Any, ClassVar
    from ...config.types import SchedulerConfigDescriptor, SLURMPartitionsDescriptor
    from ...core.element import ElementResources
    from ..jobscript import Jobscript
    from ..types import VersionInfo
    from ..shells.base import Shell


@hydrate
class SlurmPosix(QueuedScheduler):
    """
    A scheduler that uses SLURM.

    Keyword Args
    ------------
    directives: dict
        Scheduler directives. Each item is written verbatim in the jobscript as a
        scheduler directive, and is not processed in any way. If a value is `None`, the
        key is considered a flag-like directive. If a value is a list, multiple directives
        will be printed to the jobscript with the same key, but different values.

    Notes
    -----
    - runs in current working directory by default [2]

    Todo
    ----
    - consider getting memory usage like: https://stackoverflow.com/a/44143229/5042280

    References
    ----------
    [1] https://manpages.org/sbatch
    [2] https://ri.itservices.manchester.ac.uk/csf4/batch/sge-to-slurm/

    """

    #: Default submission command.
    DEFAULT_SUBMIT_CMD: ClassVar[str] = "sbatch"
    #: Default command to show the queue state.
    DEFAULT_SHOW_CMD: ClassVar[Sequence[str]] = ("squeue", "--me")
    #: Default cancel command.
    DEFAULT_DEL_CMD: ClassVar[str] = "scancel"
    #: Default job control directive prefix.
    DEFAULT_JS_CMD: ClassVar[str] = "#SBATCH"
    #: Default prefix to enable array processing.
    DEFAULT_ARRAY_SWITCH: ClassVar[str] = "--array"
    #: Default shell variable with array ID.
    DEFAULT_ARRAY_ITEM_VAR: ClassVar[str] = "SLURM_ARRAY_TASK_ID"
    #: Number of times to try when querying the state.
    NUM_STATE_QUERY_TRIES: ClassVar[int] = 5
    #: Delay (in seconds) between attempts to query the state.
    INTER_STATE_QUERY_DELAY: ClassVar[float] = 0.5

    #: Maps scheduler state codes to :py:class:`JobscriptElementState` values.
    state_lookup: ClassVar[Mapping[str, JobscriptElementState]] = {
        "PENDING": JobscriptElementState.pending,
        "RUNNING": JobscriptElementState.running,
        "COMPLETING": JobscriptElementState.running,
        "CANCELLED": JobscriptElementState.cancelled,
        "COMPLETED": JobscriptElementState.finished,
        "FAILED": JobscriptElementState.errored,
        "OUT_OF_MEMORY": JobscriptElementState.errored,
        "TIMEOUT": JobscriptElementState.errored,
    }

    @classmethod
    @override
    @TimeIt.decorator
    def process_resources(
        cls, resources: ElementResources, scheduler_config: SchedulerConfigDescriptor
    ) -> None:
        """Perform scheduler-specific processing to the element resources.

        Note
        ----
        This mutates `resources`.
        """
        if resources.is_parallel:
            if resources.parallel_mode is None:
                # set default parallel mode:
                resources.parallel_mode = ParallelMode.DISTRIBUTED

            if resources.parallel_mode is ParallelMode.SHARED:
                if (resources.num_nodes and resources.num_nodes > 1) or (
                    resources.SLURM_num_nodes and resources.SLURM_num_nodes > 1
                ):
                    raise IncompatibleParallelModeError(resources.parallel_mode)
                # consider `num_cores` and `num_threads` synonyms in this case:
                if resources.SLURM_num_tasks and resources.SLURM_num_tasks != 1:
                    raise IncompatibleSLURMArgumentsError(
                        f"For the {resources.parallel_mode.name.lower()} parallel mode, "
                        f"`SLURM_num_tasks` must be set to 1 (to ensure all requested "
                        f"cores reside on the same node)."
                    )
                resources.SLURM_num_tasks = 1

                if resources.SLURM_num_cpus_per_task == 1:
                    raise IncompatibleSLURMArgumentsError(
                        f"For the {resources.parallel_mode.name.lower()} parallel mode, "
                        f"if `SLURM_num_cpus_per_task` is set, it must be set to the "
                        f"number of threads/cores to use, and so must be greater than 1, "
                        f"but {resources.SLURM_num_cpus_per_task!r} was specified."
                    )
                resources.num_threads = resources.num_threads or resources.num_cores
                if not resources.num_threads and not resources.SLURM_num_cpus_per_task:
                    raise ValueError(
                        f"For the {resources.parallel_mode.name.lower()} parallel "
                        f"mode, specify `num_threads` (or its synonym for this "
                        f"parallel mode: `num_cores`), or the SLURM-specific "
                        f"parameter `SLURM_num_cpus_per_task`."
                    )
                elif (resources.num_threads and resources.SLURM_num_cpus_per_task) and (
                    resources.num_threads != resources.SLURM_num_cpus_per_task
                ):
                    raise IncompatibleSLURMArgumentsError(
                        f"Incompatible parameters for `num_cores`/`num_threads` "
                        f"({resources.num_threads}) and `SLURM_num_cpus_per_task` "
                        f"({resources.SLURM_num_cpus_per_task}) for the "
                        f"{resources.parallel_mode.name.lower()} parallel mode."
                    )
                resources.SLURM_num_cpus_per_task = resources.num_threads

            elif resources.parallel_mode is ParallelMode.DISTRIBUTED:
                if resources.num_threads:
                    raise ValueError(
                        f"For the {resources.parallel_mode.name.lower()} parallel "
                        f"mode, specifying `num_threads` is not permitted."
                    )
                if (
                    resources.SLURM_num_tasks
                    and resources.num_cores
                    and resources.SLURM_num_tasks != resources.num_cores
                ):
                    raise IncompatibleSLURMArgumentsError(
                        f"Incompatible parameters for `num_cores` ({resources.num_cores})"
                        f" and `SLURM_num_tasks` ({resources.SLURM_num_tasks}) for the "
                        f"{resources.parallel_mode.name.lower()} parallel mode."
                    )
                elif not resources.SLURM_num_tasks and resources.num_cores:
                    resources.SLURM_num_tasks = resources.num_cores
                elif (
                    resources.SLURM_num_tasks_per_node
                    and resources.num_cores_per_node
                    and resources.SLURM_num_tasks_per_node != resources.num_cores_per_node
                ):
                    raise IncompatibleSLURMArgumentsError(
                        f"Incompatible parameters for `num_cores_per_node` "
                        f"({resources.num_cores_per_node}) and `SLURM_num_tasks_per_node`"
                        f" ({resources.SLURM_num_tasks_per_node}) for the "
                        f"{resources.parallel_mode.name.lower()} parallel mode."
                    )
                elif (
                    not resources.SLURM_num_tasks_per_node
                    and resources.num_cores_per_node
                ):
                    resources.SLURM_num_tasks_per_node = resources.num_cores_per_node

                if (
                    resources.SLURM_num_nodes
                    and resources.num_nodes
                    and resources.SLURM_num_nodes != resources.num_nodes
                ):
                    raise IncompatibleSLURMArgumentsError(
                        f"Incompatible parameters for `num_nodes` ({resources.num_nodes})"
                        f" and `SLURM_num_nodes` ({resources.SLURM_num_nodes}) for the "
                        f"{resources.parallel_mode.name.lower()} parallel mode."
                    )
                elif not resources.SLURM_num_nodes and resources.num_nodes:
                    resources.SLURM_num_nodes = resources.num_nodes

            elif resources.parallel_mode is ParallelMode.HYBRID:
                raise NotImplementedError("hybrid parallel mode not yet supported.")

        else:
            if resources.SLURM_is_parallel:
                raise IncompatibleSLURMArgumentsError(
                    "Some specified SLURM-specific arguments (which indicate a parallel "
                    "job) conflict with the scheduler-agnostic arguments (which "
                    "indicate a serial job)."
                )
            if not resources.SLURM_num_tasks:
                resources.SLURM_num_tasks = 1

            if resources.SLURM_num_tasks_per_node:
                resources.SLURM_num_tasks_per_node = None

            if not resources.SLURM_num_nodes:
                resources.SLURM_num_nodes = 1

            if not resources.SLURM_num_cpus_per_task:
                resources.SLURM_num_cpus_per_task = 1

        num_cores = resources.num_cores or resources.SLURM_num_tasks
        num_cores_per_node = (
            resources.num_cores_per_node or resources.SLURM_num_tasks_per_node
        )
        num_nodes = resources.num_nodes or resources.SLURM_num_nodes
        para_mode = resources.parallel_mode

        # select matching partition if possible:
        all_parts = scheduler_config.get("partitions", {})
        if resources.SLURM_partition is not None:
            # check user-specified partition is valid and compatible with requested
            # cores/nodes:
            try:
                part = all_parts[resources.SLURM_partition]
            except KeyError:
                raise UnknownSLURMPartitionError(resources.SLURM_partition, all_parts)
            # TODO: we when we support ParallelMode.HYBRID, these checks will have to
            # consider the total number of cores requested per node
            # (num_cores_per_node * num_threads)?
            part_num_cores = part.get("num_cores", ())
            part_num_cores_per_node = part.get("num_cores_per_node", ())
            part_num_nodes = part.get("num_nodes", ())
            part_para_modes = part.get("parallel_modes", ())
            if cls.__is_present_unsupported(num_cores, part_num_cores):
                raise IncompatibleSLURMPartitionError(
                    resources.SLURM_partition, "number of cores", num_cores
                )
            if cls.__is_present_unsupported(num_cores_per_node, part_num_cores_per_node):
                raise IncompatibleSLURMPartitionError(
                    resources.SLURM_partition,
                    "number of cores per node",
                    num_cores_per_node,
                )
            if cls.__is_present_unsupported(num_nodes, part_num_nodes):
                raise IncompatibleSLURMPartitionError(
                    resources.SLURM_partition, "number of nodes", num_nodes
                )
            if para_mode and para_mode.name.lower() not in part_para_modes:
                raise IncompatibleSLURMPartitionError(
                    resources.SLURM_partition, "parallel mode", para_mode
                )
        else:
            # find the first compatible partition if one exists:
            # TODO: bug here? not finding correct partition?
            for part_name, part_info in all_parts.items():
                if cls.__partition_matches(
                    num_cores, num_cores_per_node, num_nodes, para_mode, part_info
                ):
                    resources.SLURM_partition = str(part_name)
                    break

    @classmethod
    def __is_present_unsupported(
        cls, num_req: int | None, part_have: Sequence[int] | None
    ) -> bool:
        """
        Test if information is present on both sides, but doesn't match.
        """
        return bool(
            num_req and part_have and not cls.is_num_cores_supported(num_req, part_have)
        )

    @classmethod
    def __is_present_supported(
        cls, num_req: int | None, part_have: Sequence[int] | None
    ) -> bool:
        """
        Test if information is present on both sides, and also matches.
        """
        return bool(
            num_req and part_have and cls.is_num_cores_supported(num_req, part_have)
        )

    @classmethod
    def __partition_matches(
        cls,
        num_cores: int | None,
        num_cores_per_node: int | None,
        num_nodes: int | None,
        para_mode: ParallelMode | None,
        part_info: SLURMPartitionsDescriptor,
    ) -> bool:
        """
        Check whether a partition (part_name, part_info) matches the requested number
        of cores and nodes.
        """
        part_num_cores = part_info.get("num_cores", [])
        part_num_cores_per_node = part_info.get("num_cores_per_node", [])
        part_num_nodes = part_info.get("num_nodes", [])
        part_para_modes = part_info.get("parallel_modes", [])
        if (
            not cls.__is_present_supported(num_cores, part_num_cores)
            or not cls.__is_present_supported(num_cores_per_node, part_num_cores_per_node)
            or not cls.__is_present_supported(num_nodes, part_num_nodes)
        ):
            return False
        # FIXME: Does the next check come above or below the check below?
        # Surely not both!
        part_match = True
        if part_match:
            return True
        if para_mode and para_mode.name.lower() not in part_para_modes:
            return False
        if part_match:
            return True
        return False

    def __format_core_request_lines(self, resources: ElementResources) -> Iterator[str]:
        if resources.SLURM_partition:
            yield f"{self.js_cmd} --partition {resources.SLURM_partition}"
        if resources.SLURM_num_nodes:  # TODO: option for --exclusive ?
            yield f"{self.js_cmd} --nodes {resources.SLURM_num_nodes}"
        if resources.SLURM_num_tasks:
            yield f"{self.js_cmd} --ntasks {resources.SLURM_num_tasks}"
        if resources.SLURM_num_tasks_per_node:
            yield f"{self.js_cmd} --ntasks-per-node {resources.SLURM_num_tasks_per_node}"
        if resources.SLURM_num_cpus_per_task:
            yield f"{self.js_cmd} --cpus-per-task {resources.SLURM_num_cpus_per_task}"

    def __format_array_request(self, num_elements: int, resources: ElementResources):
        # TODO: Slurm docs start indices at zero, why are we starting at one?
        #   https://slurm.schedmd.com/sbatch.html#OPT_array
        max_str = f"%{resources.max_array_items}" if resources.max_array_items else ""
        return f"{self.js_cmd} {self.array_switch} 1-{num_elements}{max_str}"

    def get_stdout_filename(
        self, js_idx: int, job_ID: str, array_idx: int | None = None
    ) -> str:
        """File name of the standard output stream file."""
        array_idx_str = f".{array_idx}" if array_idx is not None else ""
        return f"js_{js_idx}.sh_{job_ID}{array_idx_str}.out"

    def get_stderr_filename(
        self, js_idx: int, job_ID: str, array_idx: int | None = None
    ) -> str:
        """File name of the standard error stream file."""
        array_idx_str = f".{array_idx}" if array_idx is not None else ""
        return f"js_{js_idx}.sh_{job_ID}{array_idx_str}.err"

    def __format_std_stream_file_option_lines(
        self, is_array: bool, sub_idx: int, js_idx: int, combine_std: bool
    ) -> Iterator[str]:
        pattern = R"%x_%A.%a" if is_array else R"%x_%j"
        base = f"./artifacts/submissions/{sub_idx}/js_std/{js_idx}/{pattern}"
        yield f"{self.js_cmd} --output {base}.out"
        if not combine_std:
            yield f"{self.js_cmd} --error {base}.err"

    @override
    def format_directives(
        self,
        resources: ElementResources,
        num_elements: int,
        is_array: bool,
        sub_idx: int,
        js_idx: int,
    ) -> str:
        """
        Format the directives to the scheduler.
        """
        opts: list[str] = []
        opts.extend(self.__format_core_request_lines(resources))

        if is_array:
            opts.append(self.__format_array_request(num_elements, resources))

        opts.extend(
            self.__format_std_stream_file_option_lines(
                is_array, sub_idx, js_idx, resources.combine_jobscript_std
            )
        )

        for opt_k, opt_v in self.directives.items():
            if isinstance(opt_v, list):
                for i in opt_v:
                    opts.append(f"{self.js_cmd} {opt_k} {i}")
            elif opt_v:
                opts.append(f"{self.js_cmd} {opt_k} {opt_v}")
            elif opt_v is None:
                opts.append(f"{self.js_cmd} {opt_k}")

        return "\n".join(opts) + "\n"

    @override
    @TimeIt.decorator
    def get_version_info(self) -> VersionInfo:
        vers_cmd = [self.submit_cmd, "--version"]
        proc = subprocess.run(
            args=vers_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout = proc.stdout.decode().strip()
        stderr = proc.stderr.decode().strip()
        if stderr:
            print(stderr)
        name, version = stdout.split()
        return {
            "scheduler_name": name,
            "scheduler_version": version,
        }

    @override
    def get_submit_command(
        self,
        shell: Shell,
        js_path: str,
        deps: dict[Any, tuple[Any, ...]],
    ) -> list[str]:
        """
        Get the command to use to submit a job to the scheduler.

        Returns
        -------
        List of argument words.
        """
        cmd = [self.submit_cmd, "--parsable"]
        if deps:
            cmd.append("--dependency")
            cmd.append(",".join(self.__dependency_args(deps)))
        cmd.append(js_path)
        return cmd

    @staticmethod
    def __dependency_args(deps: dict[Any, tuple[Any, ...]]) -> Iterator[str]:
        for job_ID, is_array_dep in deps.values():
            if is_array_dep:  # array dependency
                yield f"aftercorr:{job_ID}"
            else:
                yield f"afterany:{job_ID}"

    def parse_submission_output(self, stdout: str) -> str:
        """Extract scheduler reference for a newly submitted jobscript"""
        if ";" in stdout:
            return stdout.split(";")[0]  # since we submit with "--parsable"
        # Try using the whole thing
        return stdout

    @staticmethod
    def _parse_job_IDs(job_ID_str: str) -> tuple[str, None | list[int]]:
        """
        Parse the job ID column from the `squeue` command (the `%i` format option).

        Returns
        -------
        job_id
            The job identifier.
        array_indices
            The indices into the job array.
        """
        base_job_ID, *arr_idx_data = job_ID_str.split("_")
        if not arr_idx_data:
            return base_job_ID, None
        arr_idx = arr_idx_data[0]
        try:
            return base_job_ID, [int(arr_idx) - 1]  # zero-index
        except ValueError:
            pass
        # split on commas (e.g. "[5,8-40]")
        _arr_idx: list[int] = []
        for i_range_str in arr_idx.strip("[]").split(","):
            if "-" in i_range_str:
                _from, _to = i_range_str.split("-")
                if "%" in _to:
                    # indicates max concurrent array items; not needed
                    _to = _to.split("%")[0]
                _arr_idx.extend(range(int(_from) - 1, int(_to)))
            else:
                _arr_idx.append(int(i_range_str) - 1)
        return base_job_ID, _arr_idx

    def __parse_job_states(
        self, stdout: str
    ) -> dict[str, JobscriptElementState | dict[int, JobscriptElementState]]:
        """Parse output from Slurm `squeue` command with a simple format."""
        info: dict[str, JobscriptElementState | dict[int, JobscriptElementState]] = {}
        for ln in stdout.split("\n"):
            if not ln:
                continue
            job_id, job_state, *_ = ln.split()
            base_job_ID, arr_idx = self._parse_job_IDs(job_id)
            state = self.state_lookup.get(job_state, JobscriptElementState.errored)

            if arr_idx is not None:
                entry = cast(
                    dict[int, JobscriptElementState], info.setdefault(base_job_ID, {})
                )
                for arr_idx_i in arr_idx:
                    entry[arr_idx_i] = state
            else:
                info[base_job_ID] = state

        return info

    def __query_job_states(self, job_IDs: Iterable[str]) -> tuple[str, str]:
        """Query the state of the specified jobs."""
        cmd = [
            *self.show_cmd,
            "--noheader",
            "--format",
            R"%200i %30T",  # job ID (<base_job_id>_<index> for array job) and job state
            "--jobs",
            ",".join(job_IDs),
        ]
        return run_cmd(cmd, logger=self._app.submission_logger)

    def __get_job_valid_IDs(self, job_IDs: Collection[str] | None = None) -> set[str]:
        """Get a list of job IDs that are known by the scheduler, optionally filtered by
        specified job IDs."""

        cmd = [*self.show_cmd, "--noheader", "--format", r"%F"]
        stdout, stderr = run_cmd(cmd, logger=self._app.submission_logger)
        if stderr:
            raise ValueError(
                f"Could not get query Slurm jobs. Command was: {cmd!r}; stderr was: "
                f"{stderr}"
            )
        else:
            known_jobs = set(i.strip() for i in stdout.split("\n") if i.strip())
        if job_IDs is None:
            return known_jobs
        return known_jobs.intersection(job_IDs)

    @override
    def get_job_state_info(
        self, *, js_refs: Sequence[str] | None = None
    ) -> Mapping[str, JobscriptElementState | Mapping[int, JobscriptElementState]]:
        """Query the scheduler to get the states of all of this user's jobs, optionally
        filtering by specified job IDs.

        Jobs that are not in the scheduler's status output will not appear in the output
        of this method.
        """

        # if job_IDs are passed, then assume they are existant, otherwise retrieve valid
        # jobs:
        refs: Collection[str] = js_refs or self.__get_job_valid_IDs()

        count = 0
        while refs:
            stdout, stderr = self.__query_job_states(refs)
            if not stderr:
                return self.__parse_job_states(stdout)
            if (
                "Invalid job id specified" not in stderr
                or count >= self.NUM_STATE_QUERY_TRIES
            ):
                raise ValueError(f"Could not get Slurm job states. Stderr was: {stderr}")

            # the job might have finished; this only seems to happen if a single
            # non-existant job ID is specified; for multiple non-existant jobs, no
            # error is produced;
            self._app.submission_logger.info(
                "A specified job ID is non-existant; refreshing known job IDs..."
            )
            time.sleep(self.INTER_STATE_QUERY_DELAY)
            refs = self.__get_job_valid_IDs(refs)
            count += 1
        return {}

    @override
    def cancel_jobs(
        self,
        js_refs: list[str],
        jobscripts: list[Jobscript] | None = None,
    ):
        """
        Cancel submitted jobs.
        """
        cmd = [self.del_cmd, *js_refs]
        self._app.submission_logger.info(
            f"cancelling {self.__class__.__name__} jobscripts with command: {cmd}."
        )
        stdout, stderr = run_cmd(cmd, logger=self._app.submission_logger)
        if stderr:
            raise ValueError(
                f"Could not get query {self.__class__.__name__} jobs. Command was: "
                f"{cmd!r}; stderr was: {stderr}"
            )
        self._app.submission_logger.info(
            f"jobscripts cancel command executed; stdout was: {stdout}."
        )



from __future__ import annotations
from pathlib import Path
from typing import Any, Sequence, TYPE_CHECKING
import numpy as np
import pytest
from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import (
    P1_parameter_cls as P1,
    P1_sub_parameter_cls as P1_sub,
    command_line_test,
)


def test_get_command_line(null_config, tmp_path: Path):
    p1_value = 1
    cmd_str = "Write-Output (<<parameter:p1>> + 100)"
    expected = f"Write-Output ({p1_value} + 100)"
    command_line_test(cmd_str, expected, {"p1": p1_value}, tmp_path)


@pytest.mark.parametrize("shell_args", [("powershell", "nt"), ("bash", "posix")])
def test_get_command_line_with_stdout(
    null_config, tmp_path: Path, shell_args: tuple[str, str]
):
    p1_value = 1
    expected = {
        ("powershell", "nt"): f"$parameter_p2 = Write-Output ({p1_value} + 100)",
        ("bash", "posix"): f"parameter_p2=`Write-Output ({p1_value} + 100)`",
    }
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1>> + 100)",
        cmd_stdout="<<parameter:p2>>",
        expected=expected[shell_args],
        inputs={"p1": p1_value},
        outputs=("p2",),
        shell_args=shell_args,
        path=tmp_path,
    )


def test_get_command_line_single_labelled_input(null_config, tmp_path: Path):
    p1_value = 1
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1[one]>> + 100)",
        expected=f"Write-Output ({p1_value} + 100)",
        schema_inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"), labels={"one": {}})],
        inputs=[hf.InputValue("p1", label="one", value=p1_value)],
        path=tmp_path,
    )


def test_get_command_line_multiple_labelled_input(null_config, tmp_path: Path):
    p1_one_value = 1
    p1_two_value = 2
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1[one]>> + <<parameter:p1[two]>> + 100)",
        expected=f"Write-Output ({p1_one_value} + {p1_two_value} + 100)",
        schema_inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"), multiple=True, labels={"one": {}, "two": {}}
            )
        ],
        inputs=[
            hf.InputValue("p1", label="one", value=p1_one_value),
            hf.InputValue("p1", label="two", value=p1_two_value),
        ],
        path=tmp_path,
    )


def test_get_command_line_sub_parameter(null_config, tmp_path: Path):
    p1_value = {"a": 1}
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1.a>> + 100)",
        expected=f"Write-Output ({p1_value['a']} + 100)",
        inputs={"p1": p1_value},
        path=tmp_path,
    )


def test_get_command_line_sum(null_config, tmp_path: Path):
    p1_value = [1, 2, 3]
    command_line_test(
        cmd_str="Write-Output (<<sum(parameter:p1)>> + 100)",
        expected=f"Write-Output ({sum(p1_value)} + 100)",
        inputs={"p1": p1_value},
        path=tmp_path,
    )


def test_get_command_line_join(null_config, tmp_path: Path):
    p1_value = [1, 2, 3]
    delim = ","
    command_line_test(
        cmd_str=f'Write-Output (<<join[delim="{delim}"](parameter:p1)>> + 100)',
        expected=f"Write-Output ({delim.join(str(i) for i in p1_value)} + 100)",
        inputs={"p1": p1_value},
        path=tmp_path,
    )


def test_get_command_line_sum_sub_data(null_config, tmp_path: Path):
    p1_value = {"a": [1, 2, 3]}
    command_line_test(
        cmd_str="Write-Output (<<sum(parameter:p1.a)>> + 100)",
        expected=f"Write-Output ({sum(p1_value['a'])} + 100)",
        inputs={"p1": p1_value},
        path=tmp_path,
    )


def test_get_command_line_join_sub_data(null_config, tmp_path):
    delim = ","
    p1_value = {"a": [1, 2, 3]}
    command_line_test(
        cmd_str=f'Write-Output (<<join[delim="{delim}"](parameter:p1.a)>> + 100)',
        expected=f"Write-Output ({delim.join(str(i) for i in p1_value['a'])} + 100)",
        inputs={"p1": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value(null_config, tmp_path: Path):
    p1_value = P1(a=1)  # has a `CLI_format` method defined which returns `str(a)`
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1c>> + 100)",
        expected=f"Write-Output ({p1_value.a} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_join(null_config, tmp_path: Path):
    delim = ","
    p1_value = P1(a=4)
    command_line_test(
        cmd_str=(
            f"Write-Output "
            f'<<join[delim="{delim}"](parameter:p1c.custom_CLI_format_prep(reps=4))>>'
        ),
        expected="Write-Output 4,4,4,4",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_custom_method(null_config, tmp_path: Path):
    p1_value = P1(a=1)
    command_line_test(
        cmd_str="Write-Output (<<parameter:p1c.custom_CLI_format()>> + 100)",
        expected=f"Write-Output ({p1_value.a + 4} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_custom_method_with_args(
    null_config, tmp_path: Path
):
    p1_value = P1(a=1)
    add_val = 35
    command_line_test(
        cmd_str=f"Write-Output (<<parameter:p1c.custom_CLI_format(add={add_val})>> + 100)",
        expected=f"Write-Output ({p1_value.a + add_val} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_custom_method_with_two_args(
    null_config, tmp_path: Path
):
    add_val = 35
    sub_val = 10
    p1_value = P1(a=1)
    command_line_test(
        cmd_str=(
            f"Write-Output ("
            f"<<parameter:p1c.custom_CLI_format(add={add_val}, sub={sub_val})>> + 100)"
        ),
        expected=f"Write-Output ({p1_value.a + add_val - sub_val} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_sub_object(null_config, tmp_path: Path):
    p1_value = P1(a=1, sub_param=P1_sub(e=5))
    assert p1_value.sub_param
    command_line_test(
        cmd_str=f"Write-Output (<<parameter:p1c.sub_param>> + 100)",
        expected=f"Write-Output ({p1_value.sub_param.e} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_get_command_line_parameter_value_sub_object_attr(null_config, tmp_path: Path):
    p1_value = P1(a=1, sub_param=P1_sub(e=5))
    assert p1_value.sub_param
    command_line_test(
        cmd_str=f"Write-Output (" f"<<parameter:p1c.sub_param.e>> + 100)",
        expected=f"Write-Output ({p1_value.sub_param.e} + 100)",
        inputs={"p1c": p1_value},
        path=tmp_path,
    )


def test_process_std_stream_int(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<int(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="101", stderr=False) == 101


def test_process_std_stream_stderr_int(null_config) -> None:
    cmd = hf.Command(command="", stderr="<<int(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="101", stderr=True) == 101


def test_process_std_stream_float(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<float(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="3.1415", stderr=False) == 3.1415


def test_process_std_stream_bool_true(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("true", "True", "1"):
        assert cmd.process_std_stream(name="p2", value=value, stderr=False) == True


def test_process_std_stream_bool_false(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("false", "False", "0"):
        assert cmd.process_std_stream(name="p2", value=value, stderr=False) == False


def test_process_std_stream_bool_raise(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("hi", "120", "-1"):
        with pytest.raises(ValueError):
            cmd.process_std_stream(name="p2", value=value, stderr=False)


def test_process_std_stream_list(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<list(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="1 2 3", stderr=False) == [
        "1",
        "2",
        "3",
    ]


def test_process_std_stream_list_int(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<list[item_type=int](parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="1 2 3", stderr=False) == [1, 2, 3]


def test_process_std_stream_list_delim(null_config) -> None:
    cmd = hf.Command(command="", stdout='<<list[delim=","](parameter:p2)>>')
    assert cmd.process_std_stream(name="p2", value="1,2,3", stderr=False) == [
        "1",
        "2",
        "3",
    ]


def test_process_std_stream_list_int_delim(null_config) -> None:
    cmd = hf.Command(
        command="", stdout='<<list[item_type=int, delim=","](parameter:p2)>>'
    )
    assert cmd.process_std_stream(name="p2", value="1,2,3", stderr=False) == [1, 2, 3]


def test_process_std_stream_list_float_delim_colon(null_config) -> None:
    cmd = hf.Command(
        command="", stdout='<<list[item_type=float, delim=":"](parameter:p2)>>'
    )
    assert cmd.process_std_stream(name="p2", value="1.1:2.2:3.3", stderr=False) == [
        1.1,
        2.2,
        3.3,
    ]


def test_process_std_stream_array(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<array(parameter:p2)>>")
    assert np.allclose(
        cmd.process_std_stream(name="p2", value="1 2 3", stderr=False),
        np.array([1, 2, 3]),
    )


def test_process_std_stream_array_delim(null_config) -> None:
    cmd = hf.Command(command="", stdout='<<array[delim=","](parameter:p2)>>')
    assert np.allclose(
        cmd.process_std_stream(name="p2", value="1,2,3", stderr=False),
        np.array([1, 2, 3]),
    )


def test_process_std_stream_array_dtype_int(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<array[item_type=int](parameter:p2)>>")
    arr = cmd.process_std_stream(name="p2", value="1 2 3", stderr=False)
    assert arr.dtype == np.dtype("int")


def test_process_std_stream_array_dtype_float(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<array[item_type=float](parameter:p2)>>")
    arr = cmd.process_std_stream(name="p2", value="1 2 3", stderr=False)
    assert arr.dtype == np.dtype("float")


def test_process_std_stream_object(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<parameter:p1c>>")
    a_val = 12
    assert cmd.process_std_stream(name="p1c", value=str(a_val), stderr=False) == P1(
        a=a_val
    )


def test_process_std_stream_object_kwargs(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<parameter:p1c.CLI_parse(double=true)>>")
    a_val = 12
    expected = 2 * a_val
    assert cmd.process_std_stream(name="p1c", value=str(a_val), stderr=False) == P1(
        a=expected
    )


def test_get_output_types(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<parameter:p1_test_123>>")
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_int(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<int(parameter:p1_test_123)>>")
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_object_with_args(null_config) -> None:
    cmd = hf.Command(
        command="", stdout="<<parameter:p1_test_123.CLI_parse(double=true)>>"
    )
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_list(null_config) -> None:
    cmd = hf.Command(
        command="", stdout="<<list[item_type=int, delim=" "](parameter:p1_test_123)>>"
    )
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_no_match(null_config) -> None:
    cmd = hf.Command(command="", stdout="parameter:p1_test_123")
    assert cmd.get_output_types() == {"stdout": None, "stderr": None}


def test_get_output_types_raise_with_extra_substring_start(null_config) -> None:
    cmd = hf.Command(command="", stdout="hello: <<parameter:p1_test_123>>")
    with pytest.raises(ValueError):
        cmd.get_output_types()


def test_get_output_types_raise_with_extra_substring_end(null_config) -> None:
    cmd = hf.Command(command="", stdout="<<parameter:p1_test_123>> hello")
    with pytest.raises(ValueError):
        cmd.get_output_types()


def test_extract_executable_labels(null_config) -> None:
    tests = {
        "<<executable:m1>> and <<executable:12>>": ["m1", "12"],
        "<<executable:m1>> hi": ["m1"],
        "<<executable:m1": [],
    }
    for k, v in tests.items():
        assert hf.Command._extract_executable_labels(k) == v



def get_hash(obj):
    """Return a hash from an arbitrarily nested structure dicts, lists, tuples, and
    sets.

    Note the resulting hash is not necessarily stable across sessions or machines.
    """

    if isinstance(obj, (set, tuple, list)):
        return hash(tuple([type(obj)] + [get_hash(i) for i in obj]))

    elif not isinstance(obj, dict):
        return hash(obj)

    new_obj = {k: get_hash(obj[k]) for k in obj}

    return hash(frozenset(sorted(new_obj.items())))



"""
Workflow persistence subsystem.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import PersistentStore


def store_cls_from_str(store_format: str) -> type[PersistentStore]:
    """
    Get the class that implements the persistence store from its name.
    """
    from .discovery import store_cls_from_str as impl

    return impl(store_format)
